
/*
 * xanim_arm.c
 *
 * Copyright (C) 1995 by Mark Podlipec. 
 * All rights reserved.
 *
 * This software may be freely copied, modified and redistributed without
 * fee for non-commerical purposes provided that this copyright notice is
 * preserved intact on all copies and modified copies.
 * 
 * There is no warranty or other guarantee of fitness of this software.
 * It is provided solely "as is". The author(s) disclaim(s) all
 * responsibility and liability with respect to this software's usage
 * or its effect upon hardware or computer systems.
 *
 */
/* The following copyright applies to all Ultimotion Segments of the Code:
 *
 * "Copyright International Business Machines Corporation 1994, All rights
 *  reserved. This product uses Ultimotion(tm) IBM video technology."
 *
 */

/*******************************
 * Revision
 *
 *
 ********************************/


#include "xanim_arm.h" 

LONG Is_ARM_File();
ULONG ARM_Read_File();
ARM_FRAME *ARM_Add_Frame();
void ARM_Free_Frame_List();
void ACT_Setup_Delta();
ULONG ARM_Read_Header();
ULONG ARM_Read_Index();
void ARM_Read_Frame();
ULONG ARM_Get_Length();

void ARM_Free_Stuff();
void ARM_Init_Tables();


ULONG ARM_Decode_MLINES();
void arm_yuv_to_rgb();


/* CODEC ROUTINES */
extern void yuv_to_rgb();
extern void QT_Gen_YUV_Tabs();
extern ULONG QT_Get_Color24();

void CMAP_Cache_Clear();
void CMAP_Cache_Init();

XA_ACTION *ACT_Get_Action();
XA_CHDR *ACT_Get_CMAP();
XA_CHDR *CMAP_Create_332();
XA_CHDR *CMAP_Create_422();
XA_CHDR *CMAP_Create_Gray();
void ACT_Add_CHDR_To_Action();
void ACT_Setup_Mapped();
void ACT_Get_CCMAP();
XA_CHDR *CMAP_Create_CHDR_From_True();
ULONG CMAP_Find_Closest();
UBYTE *UTIL_RGB_To_FS_Map();
UBYTE *UTIL_RGB_To_Map();

ULONG UTIL_Get_MSB_Long();
ULONG UTIL_Get_MSB_UShort();
void  UTIL_FPS_2_Time();
extern XA_ANIM_SETUP *XA_Get_Anim_Setup();
void XA_Free_Anim_Setup();

UBYTE *ARM_prev_buff = 0;
ULONG ARM_prev_buff_size = 0;
LONG *ARM_MOVL_X = 0;
LONG *ARM_MOVL_Y = 0;

ULONG arm_audio_attempt;
ULONG arm_audio_type;
ULONG arm_audio_freq;
ULONG arm_audio_chans;
ULONG arm_audio_bps;
ULONG XA_Add_Sound();


ULONG arm_frame_cnt;
ARM_FRAME *arm_frame_start,*arm_frame_cur;

ARM_FRAME *ARM_Add_Frame(time,timelo,act)
ULONG time,timelo;
XA_ACTION *act;
{
  ARM_FRAME *fframe;
 
  fframe = (ARM_FRAME *) malloc(sizeof(ARM_FRAME));
  if (fframe == 0) TheEnd1("ARM_Add_Frame: malloc err");
 
  fframe->time   = time;
  fframe->timelo = timelo;
  fframe->act = act;
  fframe->next = 0;
 
  if (arm_frame_start == 0) arm_frame_start = fframe;
  else arm_frame_cur->next = fframe;
 
  arm_frame_cur = fframe;
  arm_frame_cnt++;
  return(fframe);
}

void ARM_Free_Frame_List(fframes)
ARM_FRAME *fframes;
{
  ARM_FRAME *ftmp;
  while(fframes != 0)
  {
    ftmp = fframes;
    fframes = fframes->next;
    MYFREE(ftmp,0xA000);
  }
}

#define ARM_ARMo 0x41524D6F
#define ARM_vie  0x76696500


/*
 *
 */
LONG Is_ARM_File(filename)
char *filename;
{
  FILE *fin;
  ULONG data0,data1;

  if ( (fin=fopen(filename,XA_OPEN_MODE)) == 0) return(XA_NOFILE);
  data0 = UTIL_Get_MSB_Long(fin);
  data1 = (UTIL_Get_MSB_Long(fin)) & 0xffffff00;
 
  fclose(fin);
  if ( (data0 == ARM_ARMo) && (data1 == ARM_vie)) return(TRUE);
  return(FALSE);
}


ULONG ARM_Read_File(fname,anim_hdr,audio_attempt)
char *fname;
XA_ANIM_HDR *anim_hdr;
ULONG audio_attempt;	/* TRUE if audio is to be attempted */
{
  FILE *fin;
  LONG i,t_time;
  ULONG t_timelo;
  XA_ACTION *act;
  XA_ANIM_SETUP *arm;
  ARM_HDR   arm_hdr;
 
  if ( (fin=fopen(fname,XA_OPEN_MODE)) == 0)
  {
    fprintf(stderr,"can't open ARM File %s for reading\n",fname);
    return(FALSE);
  }

  arm = XA_Get_Anim_Setup();
  arm->vid_time = XA_GET_TIME( 100 ); /* default */

  arm_frame_cnt		= 0;
  arm_frame_start	= 0;
  arm_frame_cur		= 0;
  arm_audio_attempt	= audio_attempt;

  if (ARM_Read_Header(fin,arm,&arm_hdr) == FALSE)
  {
    fprintf(stderr,"ARM: read header error\n");
    fclose(fin);
    return(FALSE);
  }

  if (ARM_Read_Index(fin,anim_hdr,arm,&arm_hdr)==FALSE) return(FALSE);

  if (xa_verbose) 
  {
    fprintf(stderr,"ARM %ldx%ldx%ld frames %ld\n",
		arm->imagex,arm->imagey,arm->imagec,arm_frame_cnt);
  }
  if (arm_frame_cnt == 0)
  { 
    fprintf(stderr,"ARM: No supported video frames exist in this file.\n");
    return(FALSE);
  }

  anim_hdr->frame_lst = (XA_FRAME *)
				malloc( sizeof(XA_FRAME) * (arm_frame_cnt+1));
  if (anim_hdr->frame_lst == NULL) TheEnd1("ARM_Read_File: frame malloc err");

  arm_frame_cur = arm_frame_start;
  i = 0;
  t_time = 0;
  t_timelo = 0;
  while(arm_frame_cur != 0)
  {
    if (i > arm_frame_cnt)
    {
      fprintf(stderr,"ARM_Read_Anim: frame inconsistency %ld %ld\n",
                i,arm_frame_cnt);
      break;
    }
    anim_hdr->frame_lst[i].time_dur = arm_frame_cur->time;
    anim_hdr->frame_lst[i].zztime = t_time;
    t_time += arm_frame_cur->time;
    t_timelo += arm_frame_cur->timelo;
    while(t_timelo > (1<<24)) {t_time++; t_timelo -= (1<<24);}
    anim_hdr->frame_lst[i].act = arm_frame_cur->act;
    arm_frame_cur = arm_frame_cur->next;
    i++;
  }
  anim_hdr->imagex = arm->imagex;
  anim_hdr->imagey = arm->imagey;
  anim_hdr->imagec = arm->imagec;
  anim_hdr->imaged = 8; /* nop */
  anim_hdr->frame_lst[i].time_dur = 0;
  anim_hdr->frame_lst[i].zztime = -1;
  anim_hdr->frame_lst[i].act  = 0;
  anim_hdr->loop_frame = 0;
  if (xa_buffer_flag == FALSE) anim_hdr->anim_flags |= ANIM_SNG_BUF;
  if (xa_file_flag == TRUE) anim_hdr->anim_flags |= ANIM_USE_FILE;
  anim_hdr->anim_flags |= ANIM_FULL_IM;
  anim_hdr->max_fvid_size = arm->max_fvid_size;
  anim_hdr->max_faud_size = arm->max_faud_size;
  anim_hdr->fname = anim_hdr->name;
  if (i > 0) 
  {
    anim_hdr->last_frame = i - 1;
    anim_hdr->total_time = anim_hdr->frame_lst[i-1].zztime
				+ anim_hdr->frame_lst[i-1].time_dur;
  }
  else
  {
    anim_hdr->last_frame = 0;
    anim_hdr->total_time = 0;
  }
  ARM_Free_Frame_List(arm_frame_start);
  XA_Free_Anim_Setup(arm);
  return(TRUE);
} /* end of read file */


#define ARM_MAX_LINE 256
char arm_line[ARM_MAX_LINE];

ULONG ARM_Read_Line(fin,ptr)
FILE *fin;
char *ptr;
{ int i = 0;
  while( !feof(fin) )
  { int d = fgetc(fin);
    if (d >= 0)
    { ptr[i] = d; i++;
      if (i  >= ARM_MAX_LINE) return(FALSE);
      if ( (d == 0x0a) || (d == 0x0d) ) 
	{ ptr[i] = 0; 
	  DEBUG_LEVEL1 fprintf(stderr,"LINE: %s",ptr);
          return(TRUE); 
	}
    }
    else return(FALSE); 
  }
  return(FALSE);
}

ULONG ARM_Read_Header(fin,arm,arm_hdr)
FILE *fin;
XA_ANIM_SETUP *arm;
ARM_HDR *arm_hdr;
{ ULONG t;
  double tmpf;

  if (ARM_Read_Line(fin,arm_line) == FALSE) return(FALSE);  /* ARMovie */
  if (ARM_Read_Line(fin,arm_line) == FALSE) return(FALSE);  /* name */
  if (ARM_Read_Line(fin,arm_line) == FALSE) return(FALSE);  /* date/copyright */
  if (ARM_Read_Line(fin,arm_line) == FALSE) return(FALSE);  /* author/etc */
  if (ARM_Read_Line(fin,arm_line) == FALSE) return(FALSE);  /* video format */
    if (arm_line[0] == '1') arm_hdr->vid_codec = ARM_VIDEO_MOVL_RGB;
    else  arm_hdr->vid_codec = ARM_VIDEO_UNK;
  if (ARM_Read_Line(fin,arm_line) == FALSE) return(FALSE);	/* width */
    tmpf = atof(arm_line);  arm_hdr->width	 = (LONG)tmpf;
  if (ARM_Read_Line(fin,arm_line) == FALSE) return(FALSE);	/* height */
    tmpf = atof(arm_line);  arm_hdr->height	= (LONG)tmpf;
  if (ARM_Read_Line(fin,arm_line) == FALSE) return(FALSE);	/* depth */
    tmpf = atof(arm_line);  arm_hdr->depth	= (LONG)tmpf;

/* Check Video Precision Line for modifiers */
    if (arm_hdr->vid_codec == ARM_VIDEO_MOVL_RGB)
    {
       if (   (XA_find_str(arm_line,"yuv") == TRUE) 
           || (XA_find_str(arm_line,"YUV") == TRUE) )
		arm_hdr->vid_codec = ARM_VIDEO_MOVL_YUV;
    }

  if (ARM_Read_Line(fin,arm_line) == FALSE) return(FALSE);	/* fps */
    arm_hdr->fps = atof(arm_line);
  if (ARM_Read_Line(fin,arm_line) == FALSE) return(FALSE);  /*audio format */
    if (arm_audio_attempt == TRUE)
    {
      if (arm_line[0] == '0') 
      {
	arm_hdr->aud_codec = ARM_AUDIO_NONE;
	arm_audio_attempt = FALSE;
      }
      else if (arm_line[0] == '1') 
      { 
	arm_hdr->aud_codec = ARM_AUDIO_FORM1;
      }
      else
      { 
	fprintf(stderr,"ARM: Unsupported Audio Type: %s\n",arm_line);
	arm_hdr->aud_codec = ARM_AUDIO_UNK;
	arm_audio_attempt = FALSE;
      }
    }
  if (ARM_Read_Line(fin,arm_line) == FALSE) return(FALSE);  /* audio freq */
     /* NOTE: this can be fraction HZ - eventually support this or us */
    tmpf = atof(arm_line);
    { static unsigned char us[3] = { 0xB5, 0x73, 0x00 };
      if (XA_find_str(arm_line,us)==TRUE)
      { double freq = (1000000.0 / tmpf) + 0.5;
        arm_hdr->aud_freq = (LONG)freq;
      }
      else arm_hdr->aud_freq = (LONG)tmpf;
    }
     
  if (ARM_Read_Line(fin,arm_line) == FALSE) return(FALSE);  /* audio chans */
    tmpf = atof(arm_line);  arm_hdr->aud_chans	= (LONG)tmpf;
  if (ARM_Read_Line(fin,arm_line) == FALSE) return(FALSE);  /* audio prec */
    tmpf = atof(arm_line);  arm_hdr->aud_prec	= (LONG)tmpf;

    /* Check Audio Precision Line for modifiers */
    if (arm_audio_attempt == TRUE)
    {
      arm_audio_freq  = arm_hdr->aud_freq;
      arm_audio_chans = arm_hdr->aud_chans;
      if (arm_hdr->aud_prec == 8) arm_audio_bps = 1;
      else arm_audio_bps = 2;
      if ( XA_find_str(arm_line,"linear") == TRUE) 
      {
        if (xa_verbose==TRUE) 
		fprintf(stderr,"AUDIO: Signed %ld Hz %ld chans\n",
				arm_audio_freq,arm_audio_chans);
        arm_audio_type = XA_AUDIO_SIGNED;  /* NOTE: little endian */
      } 
      else if ( XA_find_str(arm_line,"logarithmic") == TRUE) 
      {
        if (xa_verbose==TRUE) 
		fprintf(stderr,"AUDIO: LOGARITHMIC %ld Hz %ld chans\n",
				arm_audio_freq,arm_audio_chans);
        arm_audio_type = XA_AUDIO_SUN_AU;  /* NOTE: little endian */
      } 
      else arm_audio_type = XA_AUDIO_INVALID;
      /* paranoia */
      if ( (arm_audio_chans > 2) || (arm_audio_chans == 0) )
					arm_audio_type = XA_AUDIO_INVALID;
      if ( (arm_audio_bps > 2) || (arm_audio_bps == 0) )
					arm_audio_type = XA_AUDIO_INVALID;

      /* warning and final modifications */
      if (arm_audio_type == XA_AUDIO_INVALID)
      {
        if (arm_audio_attempt == TRUE) 
	  fprintf(stderr,"ARM: Audio Type unsupported %ld \n",
						arm_hdr->aud_codec);
        arm_audio_attempt = FALSE;
      }
      else
      {
        if (arm_audio_chans == 2) arm_audio_type |= XA_AUDIO_STEREO_MSK;
        if (arm_audio_bps == 2) arm_audio_type |= XA_AUDIO_BPS_2_MSK;
      }
    }

  if (ARM_Read_Line(fin,arm_line) == FALSE) return(FALSE);  /*frames perchunk*/
    tmpf = atof(arm_line);  arm_hdr->fp_chunk	= (LONG)tmpf;
  if (ARM_Read_Line(fin,arm_line) == FALSE) return(FALSE);  /* chunk count */
    tmpf = atof(arm_line);  arm_hdr->chunk_cnt	= (LONG)tmpf;
  if (ARM_Read_Line(fin,arm_line) == FALSE) return(FALSE);  /*ev chunk size */
    tmpf = atof(arm_line);  arm_hdr->ev_chk_size	= (LONG)tmpf;
  if (ARM_Read_Line(fin,arm_line) == FALSE) return(FALSE);  /*od chunk size */
    tmpf = atof(arm_line);  arm_hdr->od_chk_size	= (LONG)tmpf;
  if (ARM_Read_Line(fin,arm_line) == FALSE) return(FALSE);  /* idx   offset */
    tmpf = atof(arm_line);  arm_hdr->idx_offset	= (LONG)tmpf;
  if (ARM_Read_Line(fin,arm_line) == FALSE) return(FALSE);  /* sprite offset */
    tmpf = atof(arm_line);  arm_hdr->sprite_off	= (LONG)tmpf;
  if (ARM_Read_Line(fin,arm_line) == FALSE) return(FALSE);  /* sprite size */
    tmpf = atof(arm_line);  arm_hdr->sprite_size = (LONG)tmpf;
  if (ARM_Read_Line(fin,arm_line) == FALSE) return(FALSE);  /* key  offset */
    tmpf = atof(arm_line);  arm_hdr->key_offset	= (LONG)tmpf;
  

  UTIL_FPS_2_Time(arm, arm_hdr->fps );

/*
  switch(arm_hdr->aud_codec)
  {
  }
*/

  arm->depth  = arm_hdr->depth;
  arm->imagex = arm_hdr->width;
  arm->imagey = arm_hdr->height;
  QT_Gen_YUV_Tabs();
  ARM_Init_Tables(arm->imagex,arm->imagey);

if (xa_verbose)
{
  fprintf(stderr,"ARM: %ldx%ldx%lx\n",arm->imagex,arm->imagey,arm->depth);
}

  if (   (cmap_true_map_flag == FALSE) /* depth 16 and not true_map */
      || (xa_buffer_flag == FALSE) )
  {
     if (cmap_true_to_332 == TRUE)
             arm->chdr = CMAP_Create_332(arm->cmap,&arm->imagec);
     else    arm->chdr = CMAP_Create_Gray(arm->cmap,&arm->imagec);
  }
  if ( (arm->pic==0) && (xa_buffer_flag == TRUE))
  {
    arm->pic_size = arm->imagex * arm->imagey;
    if ( (cmap_true_map_flag == TRUE) && (arm->depth > 8) )
                arm->pic = (UBYTE *) malloc( 3 * arm->pic_size );
    else arm->pic = (UBYTE *) malloc( XA_PIC_SIZE(arm->pic_size) );
    if (arm->pic == 0) TheEnd1("ARM_Buffer_Action: malloc failed");
  }
  return(TRUE);
}


ULONG ARM_Read_Index(fin,anim_hdr,arm,arm_hdr)
FILE *fin;
XA_ANIM_HDR *anim_hdr;
XA_ANIM_SETUP *arm;
ARM_HDR *arm_hdr;
{ ULONG idx_cnt = arm_hdr->chunk_cnt;
  ULONG i,*idx_off,*idx_vsize,*idx_asize;

  idx_off = (ULONG *)malloc( (idx_cnt + 2) * sizeof(ULONG) );
  idx_vsize = (ULONG *)malloc( (idx_cnt + 2) * sizeof(ULONG) );
  idx_asize = (ULONG *)malloc( (idx_cnt + 2) * sizeof(ULONG) );

  fprintf(stderr,"    idx offset  %08lx\n",arm_hdr->idx_offset);
  fseek(fin,arm_hdr->idx_offset,0);
  /* Go Through Chunks */
  for(i=0; i<=idx_cnt; i++)
  { LONG off, vid_size, aud_size, ret;
    if (ARM_Read_Line(fin,arm_line) == FALSE) return(FALSE);  /* chunk count */
    ret = sscanf(arm_line,"%ld,%ld;%ld",&off,&vid_size,&aud_size);
    DEBUG_LEVEL1
    {
     fprintf(stderr,"  (%ld) %ld) off %08lx  vid %08lx  aud %08lx\n",ret,i,
						off,vid_size,aud_size);
    }
    idx_off[i] = off;
    idx_vsize[i] = vid_size;
    idx_asize[i] = aud_size;

  } /* end of for */


  for(i=0; i<=idx_cnt; i++)
  { ULONG j; ULONG tot_dsize = 0;

    /* VIDEO */
    fseek(fin,idx_off[i],0);
    for(j=0; j < arm_hdr->fp_chunk; j++)
    { ULONG fpos,dsize;
      fpos = ftell(fin);		/* save fin position */
      dsize = ARM_Get_Length(fin,arm);	/* find length of delta */
      fseek(fin,fpos,0);		/* return to fin position */
      ARM_Read_Frame(fin,anim_hdr,arm,arm_hdr,dsize);
      tot_dsize += dsize;
    }

    /* AUDIO */
    if (arm_audio_attempt == TRUE) 
    { LONG ret, snd_size = (LONG)idx_asize[i];
      ULONG snd_off = idx_off[i] + idx_vsize[i];
fprintf(stderr,"AUD %ld) as %lx i+v %lx\n",i,idx_asize[i],snd_off);
      if (xa_file_flag==TRUE)
      { LONG rets;
	rets = XA_Add_Sound(anim_hdr,0,arm_audio_type, snd_off,
				arm_audio_freq, snd_size, 
				&arm->aud_time,&arm->aud_timelo);
	if (rets==FALSE) arm_audio_attempt = FALSE;
	if (snd_size > arm->max_faud_size) arm->max_faud_size = snd_size;
      }
      else
      { UBYTE *snd = (UBYTE *)malloc(snd_size);
	if (snd==0) TheEnd1("ARM: snd malloc err");
        fseek(fin,snd_off,0); /* seek to audio info */
	ret = fread( snd, snd_size, 1, fin);
	if (ret != 1) fprintf(stderr,"ARM: snd rd err(size %lx\n",snd_size);
	else
	{ int rets;
          /*NOTE: don't free snd */
	  rets = XA_Add_Sound(anim_hdr,snd,arm_audio_type, -1,
				arm_audio_freq, snd_size,
				&arm->aud_time, &arm->aud_timelo);
	  if (rets==FALSE) arm_audio_attempt = FALSE;
	}
      }
    } /* end of audio attempt */
  }
  free(idx_off);
  free(idx_vsize);
  free(idx_asize);
  return(TRUE);
}

/* Assuming fin is pointing to start of delta.
 * leaves fin just after end of delta.
 */
void ARM_Read_Frame(fin,anim_hdr,arm,arm_hdr,vsize)
FILE *fin;
XA_ANIM_HDR *anim_hdr;
XA_ANIM_SETUP *arm;
ARM_HDR *arm_hdr;
LONG vsize;
{ ULONG exit_flag;
  XA_ACTION *act;
  ULONG dlta_len = vsize;
  ACT_DLTA_HDR *dlta_hdr;

  act = ACT_Get_Action(anim_hdr,ACT_DELTA);
  if (xa_file_flag == TRUE)
  {
    dlta_hdr = (ACT_DLTA_HDR *) malloc(sizeof(ACT_DLTA_HDR));
    if (dlta_hdr == 0) TheEnd1("ARM: dlta malloc err");
    act->data = (UBYTE *)dlta_hdr;
    dlta_hdr->flags = ACT_SNGL_BUF;
    dlta_hdr->fsize = dlta_len;
    dlta_hdr->fpos  = ftell(fin);
    if (dlta_len > arm->max_fvid_size) arm->max_fvid_size = dlta_len;
    fseek(fin,dlta_len,1);
  }
  else
  { ULONG d; LONG ret;
    d = dlta_len + (sizeof(ACT_DLTA_HDR));
    dlta_hdr = (ACT_DLTA_HDR *) malloc( d );
    if (dlta_hdr == 0) TheEnd1("QT rle: malloc failed");
    act->data = (UBYTE *)dlta_hdr;
    dlta_hdr->flags = ACT_SNGL_BUF | DLTA_DATA;
    dlta_hdr->fpos = 0; dlta_hdr->fsize = dlta_len;
    ret = fread( dlta_hdr->data, dlta_len, 1, fin);
    if (ret != 1) { fprintf(stderr,"ARM: read err\n"); return; }
  }
  ARM_Add_Frame(arm->vid_time,arm->vid_timelo,act);
  dlta_hdr->xpos = dlta_hdr->ypos = 0;
  dlta_hdr->xsize = arm->imagex;
  dlta_hdr->ysize = arm->imagey;
  dlta_hdr->special = 0;
  if (arm_hdr->vid_codec == ARM_VIDEO_MOVL_YUV)
		dlta_hdr->extra = (void *)(ARM_VIDEO_YUV);
  else		dlta_hdr->extra = (void *)(ARM_VIDEO_RGB);
  dlta_hdr->delta = ARM_Decode_MLINES;
  ACT_Setup_Delta(arm,act,dlta_hdr,fin);
}



#define ARM_15YUV(pix,y,u,v) { \
  v = (pix >> 10) & 0x1f; u = (pix >> 5) & 0x1f; y = pix & 0x1f; \
  u ^= 0x10; v ^= 0x10; \
  y = (y << 3) | (y >> 2); u = (u << 3) | (u >> 2); v = (v << 3) | (v >> 2); \
} 

#define ARM_16YUV(pix,y,u,v) { \
  v = (pix >> 11) & 0x1f; u = (pix >> 6) & 0x1f; y = (pix >> 1) & 0x1f; \
  u ^= 0x10; v ^= 0x10; \
  y = (y << 3) | (y >> 2); u = (u << 3) | (u >> 2); v = (v << 3) | (v >> 2); \
}

#define ARM_GET_CODE(dp,code) {code = (*dp++); code |= (*dp++) << 8; }

ULONG ARM_Get_RGB_Pixel(pix,map_flag,map,chdr)
ULONG pix,map_flag,*map;
XA_CHDR *chdr;
{ ULONG r,g,b; ULONG ret;
  r = (pix >> 10) & 0x1f; g = (pix >> 5) & 0x1f; b = pix & 0x1f;
  r = (r << 3) | (r >> 2); g = (g << 3) | (g >> 2); b = (b << 3) | (b >> 2);
  return( QT_Get_Color24(r,g,b,map_flag,map,chdr) );
} 

ULONG ARM_Get_YUV_Pixel(pix,map_flag,map,chdr)
ULONG pix,map_flag,*map;
XA_CHDR *chdr;
{ ULONG y,u,v,r,g,b; ULONG ret;
  v = (pix >> 10) & 0x1f; u = (pix >> 5) & 0x1f; y = pix & 0x1f;
  u ^= 0x10; v ^= 0x10;
  y = (y << 3) | (y >> 2); u = (u << 3) | (u >> 2); v = (v << 3) | (v >> 2);
  yuv_to_rgb(y,u,v,&r,&g,&b); 
  return( QT_Get_Color24(r,g,b,map_flag,map,chdr) );
} 

/* TEST */
ULONG
ARM_Decode_MLINES(image,delta,dsize,tchdr,map,map_flag,imagex,imagey,imaged,
					xs,ys,xe,ye,special,extra)
UBYTE *image;           /* Image Buffer. */
UBYTE *delta;           /* delta data. */
ULONG dsize;            /* delta size */
XA_CHDR *tchdr;		/* color map info */
ULONG *map;             /* used if it's going to be remapped. */
ULONG map_flag;         /* whether or not to use remap_map info. */
ULONG imagex,imagey;    /* Size of image buffer. */
ULONG imaged;           /* Depth of Image. (IFF specific) */
ULONG *xs,*ys;          /* pos of changed area. */
ULONG *xe,*ye;          /* size of changed area. */
ULONG special;          /* Special Info. */
void *extra;		/* extra info needed to decode delta */
{
  UBYTE *dp = delta;
  ULONG special_flag = special & 0x0001;
  LONG i_cnt,im_size = imagex * imagey;
  ULONG code;
  XA_CHDR *chdr;
  UBYTE *iptr = image;
  UBYTE *pptr = ARM_prev_buff;
  ULONG vid_type = (ULONG)(extra);
  ULONG (*arm_get_pixel)();
  
  if (tchdr) {chdr=(tchdr->new_chdr)?(tchdr->new_chdr):(tchdr);} else chdr=0;

  /* Need to copy previous image, since Temporal references it */
  { i_cnt  = im_size * ((special_flag)?(3):(x11_bytes_pixel));
    memcpy((char *)(pptr),(char *)(image),(i_cnt));
  }

/* */
  if (vid_type == ARM_VIDEO_YUV) arm_get_pixel = ARM_Get_YUV_Pixel;
  else	arm_get_pixel = ARM_Get_RGB_Pixel;

  i_cnt = 0;
  while(i_cnt < im_size)
  { ARM_GET_CODE(dp,code); 

    if (code & 0x01)
    { ULONG test,cnt;
      if (code == 0xe601) { i_cnt = im_size; break; }	/* End of Frame */
      test = code >> 7;		cnt = ((code >> 1) & 0x3f) + 2;
      if (test < 0x1cc)			/* Temporal/Spatial */
      { UBYTE *isp;
        if (test < 0x120) { isp = pptr; isp += i_cnt; } /* temporal */
	else isp = iptr; /* spatial */
	isp += (ARM_MOVL_X[test] + (ARM_MOVL_Y[test] * imagex));
	i_cnt += cnt;	if (i_cnt >= im_size) cnt -= (i_cnt - im_size);
	while(cnt--) *iptr++ = *isp++;
      }
      else if (test >= 0x1e0)
      { cnt = ((code >> 1) & 0x3ff) + 1;
	if (test >=0x1f0)				/* New N Pixels */
	{ ULONG bs = ((cnt * 15) + 15) >> 4;
	  DEBUG_LEVEL1  fprintf(stderr,"New N Pixels %ld\n",cnt);
	  i_cnt += cnt;	if (i_cnt >= im_size) cnt -= (i_cnt - im_size);
	  iptr += cnt;  
	  dp += bs;
	  dp += bs;
	}
	else { iptr += cnt;  i_cnt += cnt; }		/* Copy/Skip */
      }
      else if (test == 0x1cc)
      { ULONG pixel; UBYTE d;
	ARM_GET_CODE(dp,pixel);
	d = (UBYTE)arm_get_pixel(pixel,map_flag,map,chdr);
	i_cnt += cnt;	if (i_cnt >= im_size) cnt -= (i_cnt - im_size);
	while(cnt--) *iptr++ = d;
      }
      else fprintf(stderr,"ARM decode ERROR: %04lx\n",code);
    }
    else					/* New Pixel */
    { 
      *iptr++ = (UBYTE)arm_get_pixel((code>>1),map_flag,map,chdr);  i_cnt++;
    }
  }
  *xs = *ys = 0; *xe = imagex; *ye = imagey;
  if (map_flag==TRUE) return(ACT_DLTA_MAPD);
  else return(ACT_DLTA_NORM);
}


ULONG ARM_Get_Length(fin,arm)
FILE *fin;
XA_ANIM_SETUP *arm;
{ ULONG d_cnt = 0;
  LONG i = arm->imagex * arm->imagey;
  ULONG code;

  while(i > 0)
  {
    if ( feof(fin) ) return(0);
    code = fgetc(fin); code |= (fgetc(fin) & 0xff) << 8; d_cnt += 2;

    if (code & 0x01)
    { ULONG test,cnt;

      if (code == 0xe601) return(d_cnt);	/* End of Frame */
      test = code >> 7;		
      cnt = ((code >> 1) & 0x3f) + 2;
      if (test < 0x120)		{ i -= cnt; }	/* Temporal */
      else if (test < 0x1cc)	{ i -= cnt; }	/* Spatial */ 
      else if (test >= 0x1e0)			/* New N Pix */
      { cnt = ((code >> 1) & 0x3ff) + 1;
	if (test >=0x1f0)			/* New N Pixels */
	{ ULONG bs = ((cnt * 15) + 15) >> 4;
	  i -= cnt;
	  while(bs--) { fgetc(fin); fgetc(fin); d_cnt += 2; }
	}
	else	{ i -= cnt; }			/* Copy/Skip */
      }
      else if (test == 0x1cc)				/* Duplicate */
      {
        if (cnt <= 2) {fprintf(stderr,"DUP CNT ERR\n"); continue; }
	fgetc(fin); fgetc(fin); d_cnt += 2;
	i -= cnt;
      }
      else 
      {
	fprintf(stderr,"ARM Get len error: %04lx\n",code);
      }
    }
    else { i--; }				/* New Pixel */
  }
  code = fgetc(fin); code |= (fgetc(fin) & 0xff) << 8;
  if (code == 0xe601) d_cnt += 2;
  return(d_cnt);
}

void arm_yuv_to_rgb(iy,iu,iv,ir,ig,ib)
ULONG iy,iu,iv;
ULONG *ir,*ig,*ib;
{ float yy,uu,vv;
  float u2,u3,v2,v3;
  float DD;
  float rr,gg,bb;
  

  DD = 255.0;
  yy =  ((float)(iy)/31.0);
  vv =  ((float)(iv)/15.0);
  uu =  ((float)(iu)/15.0);
  if (vv > 1.0) vv = -((31.0 - (float)(iv) ) / 15.0);
  if (uu > 1.0) uu = -((31.0 - (float)(iu) ) / 15.0);
  v2 = vv * 0.701;
  u2 = uu * 0.886;
  v3 = vv * (0.299 * 0.701 / 0.587);
  u3 = uu * (0.114 * 0.886 / 0.587);

  bb = (yy + u2);
  rr = (yy + v2);
  gg = (yy - u3 - v3);
  DEBUG_LEVEL1 
	fprintf(stderr,"rr = %f gg = %f bb = %f\n",rr,gg,bb);
  bb *= 255.0;
  rr *= 255.0;
  gg *= 255.0;

  if (bb > 255.0) bb = 255.0; if (bb < 0.0) bb = 0.0;
  if (rr > 255.0) rr = 255.0; if (rr < 0.0) rr = 0.0;
  if (gg > 255.0) gg = 255.0; if (gg < 0.0) gg = 0.0;
  *ir = (ULONG)(rr);
  *ig = (ULONG)(gg);
  *ib = (ULONG)(bb);

  DEBUG_LEVEL1
    fprintf(stderr,"YUV %ld %ld %ld  RGB %ld %ld %ld\n",iy,iu,iv,*ir,*ig,*ib);
}


void ARM_Init_Tables(imagex,imagey)
ULONG imagex,imagey;
{ ULONG i; LONG x,y;

 /* Prev Buffer Allocation */
 i = imagex * imagey * x11_bytes_pixel;
 if (ARM_prev_buff == 0)
 { ARM_prev_buff_size = i;
   ARM_prev_buff = (UBYTE *)malloc( ARM_prev_buff_size );
 }
 else if (i > ARM_prev_buff_size)
 { ARM_prev_buff_size = i;
   ARM_prev_buff = (UBYTE *)realloc( ARM_prev_buff, ARM_prev_buff_size );
 }
 if (ARM_prev_buff == 0) TheEnd1("ARM: prev_buff malloc err");

 /* Spatial/Temporal Tables Allocation */
 if (ARM_MOVL_X == 0)
 { ARM_MOVL_X = (LONG *)malloc( 459 * sizeof(LONG) );
   if (ARM_MOVL_X==0) TheEnd1("ARM: MOVL_X malloc err");
 }
 if (ARM_MOVL_Y == 0)
 { ARM_MOVL_Y = (LONG *)malloc( 459 * sizeof(LONG) );
   if (ARM_MOVL_Y==0) TheEnd1("ARM: MOVL_Y malloc err");
 }
 /* Temporal Encoding */
 x = y = -8;
 for(i = 0; i < 288; i++)
 { ARM_MOVL_X[i] = x;
   ARM_MOVL_Y[i] = y;
   x++; if (x > 8) { x = -8; y++; }
   if ((x==0) && (y==0))  x++; /* skip x = y = 0 case */
 }
 /* Spatial Encoding */
 x = y = -9;
 for(i = 288; i < 459; i++)
 { ARM_MOVL_X[i] = x;
   ARM_MOVL_Y[i] = y;
   x++; if (x > 9) { x = -9; y++; }
 }
}

void ARM_Free_Stuff()
{
  if (ARM_MOVL_X) { free(ARM_MOVL_X); ARM_MOVL_X = 0; }
  if (ARM_MOVL_Y) { free(ARM_MOVL_Y); ARM_MOVL_Y = 0; }
  if (ARM_prev_buff) { free(ARM_prev_buff); ARM_prev_buff = 0; }
}

